unit NewKeyMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, MPIF, MPDL;

type
  TfrmNewKey = class(TForm)
    gbNewKey: TGroupBox;
    sbStatus: TStatusBar;
    rbRSA: TRadioButton;
    gbRSA: TGroupBox;
    gbRSAPublicKey: TGroupBox;
    rbRSAe3: TRadioButton;
    rbRSAe17: TRadioButton;
    rbRSAe65537: TRadioButton;
    rbRSAeLRN: TRadioButton;
    rbRSAeSpec: TRadioButton;
    edtRSAe: TEdit;
    Label1: TLabel;
    edtRSAKeySize: TEdit;
    Label2: TLabel;
    gbRW: TGroupBox;
    Label3: TLabel;
    gbRWPublicKey: TGroupBox;
    Label4: TLabel;
    rbRWe2: TRadioButton;
    rbRWeLRN: TRadioButton;
    rbRWeSpec: TRadioButton;
    edtRWe: TEdit;
    edtRWKeySize: TEdit;
    rbRW: TRadioButton;
    rbDL: TRadioButton;
    gbDL: TGroupBox;
    rbDSA: TRadioButton;
    rbDLSpecify: TRadioButton;
    gbDLparam: TGroupBox;
    Label5: TLabel;
    edtDLpSize: TEdit;
    Label6: TLabel;
    edtDLqSize: TEdit;
    rbDLImport: TRadioButton;
    btnOpen: TButton;
    OpenDialog1: TOpenDialog;
    btnCreate: TButton;
    btnPassword: TButton;
    btnSave: TButton;
    SaveDialog1: TSaveDialog;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure rbRSAeClick(Sender: TObject);
    procedure rbRSAClick(Sender: TObject);
    procedure rbRWeClick(Sender: TObject);
    procedure rbDSAClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure btnPasswordClick(Sender: TObject);
    procedure btnCreateClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
  private
    FDLSystemParams: TDLSystemParams;
    FIFPriv: TIFPrivateKey;
    FDLPriv: TDLPrivateKey;
    procedure YarrowReseed(Sender: TObject);
  public
    { Public declarations }
  end;

var
  frmNewKey: TfrmNewKey;

implementation

uses
  SecUtils, YarrowModule, PasswordMain, MPArith, ASN1, X509, MPYarrow;

{$R *.DFM}

procedure TfrmNewKey.FormCreate(Sender: TObject);
begin

  // IMPORTANT: dmYarrow MUST NOT be auto created if this code snipplet is used:
  dmYarrow := TdmYarrow.Create(Application);
  dmYarrow.OnReseed := YarrowReseed;
  dmYarrow.OnReseeding := YarrowReseed;
end;

procedure TfrmNewKey.YarrowReseed(Sender: TObject);
begin
  sbStatus.Panels[1].Text := Format('PRNG status - Fast: %d, Slow %d, Active Threads: %d',
                                    [dmYarrow.FastReseeds,
                                     dmYarrow.SlowReseeds,
                                     dmYarrow.ReseedThreads]);
end;

var
  PW: string;

procedure TfrmNewKey.FormDestroy(Sender: TObject);
begin
  if PW <> '' then
    ProtectClear(PW[1],Length(PW));
  if Assigned(dmYarrow) then begin
    dmYarrow.OnReseed := nil;
    dmYarrow.OnReseeding := nil;
  end;
  DisposeDLSystemParams(FDLSystemParams);
  DisposeDLPrivateKey(FDLPriv);
  DisposeIFPrivateKey(FIFPriv);
end;

procedure TfrmNewKey.rbRSAeClick(Sender: TObject);
begin
  edtRSAe.Enabled := rbRSAeSpec.Checked;
end;

procedure TfrmNewKey.rbRSAClick(Sender: TObject);
begin
  gbRSA.Enabled := rbRSA.Checked;
  gbRW.Enabled := rbRW.Checked;
  gbDL.Enabled := rbDL.Checked;
end;

procedure TfrmNewKey.rbRWeClick(Sender: TObject);
begin
  edtRWe.Enabled := rbRWeSpec.Checked;
end;

procedure TfrmNewKey.rbDSAClick(Sender: TObject);
begin
  gbDLParam.Enabled := rbDLSpecify.Checked;
  btnOpen.Enabled := rbDLImport.Checked;
end;

procedure TfrmNewKey.btnOpenClick(Sender: TObject);
var
  FS: TFileStream;
  Cert: TASN1Struct;
  Publ: TDLPublicKey;
begin
  if OpenDialog1.Execute then begin
    FS := TFileStream.Create(OpenDialog1.FileName,fmOpenRead or fmShareCompat);
    try
      Cert := nil;
      try
        if ASN1FromStream(FS,Cert) = E_OK then begin
          if CheckCertSyntax(Cert) then begin
            if ExtractSubjectDHPublicKey(Cert,Publ) <> E_OK then
              if ExtractSubjectDSAPublicKey(Cert,Publ) <> E_OK then
                raise Exception.Create('Public key is not recognized as a DL key');
            DisposeDLSystemParams(FDLSystemParams);
            FDLSystemParams := Publ.Params;
            MPDealloc(Publ.Y);
            ShowMessage('Import successful');
          end else
            raise Exception.Create('Not a X.509 certificate');
        end else
          raise Exception.Create('Not a binary DER encoded file');
      finally
        Cert.Free;
      end;
    finally
      FS.Free;
    end;
  end;
end;

procedure TfrmNewKey.btnPasswordClick(Sender: TObject);
var
  N, K: packed array [0..63] of Byte;
begin
  if NewPasswordDlg('Private Key File',PW) then begin
    // The password PW will be used to seed Yarrow.
    //
    // Create seed for Yarrow:
    //  Firstly a nonce N:
    RawRandom(N,64);
    //  Secondly an iterated hash value K of PW and N:
    try
      WeakPasswordToPrivKey(PW[1],Length(PW),N,64,$1000,K,64);
    // Seed Yarrow with K:
      dmYarrow.ForceSlowReseed(K,64);
    finally
      ProtectClear(N,SizeOf(N));
      ProtectClear(K,SizeOf(K));
    end;
    // Now, hopefully, there is enough entropy in Yarrow to create a new key.
    // Enable Create Key:
    btnCreate.Enabled := True;
  end;
end;

procedure TfrmNewKey.btnCreateClick(Sender: TObject);
var
  IfPubl: TIFPublicKey;
  DLPubl: TDLPublicKey;
  KeySize: Integer;
  OS: string;
begin
  sbStatus.Panels[0].Text := 'Creating key...';
  sbStatus.Repaint;
  try
    DisposeIFPrivateKey(FIFPriv);
    DisposeDLPrivateKey(FDLPriv);

    if rbRSA.Checked then begin
      KeySize := StrToInt(edtRSAKeySize.Text);
      FillChar(IFPubl,SizeOf(IFPubl),0);
      try
        if rbRSAeLRN.Checked then begin
          MPRandom(IFPubl.E,KeySize-2);
          MPShl(IFPubl.E,1);
          IFPubl.E.Data[IFPubl.E.Size-1] := IFPubl.E.Data[IFPubl.E.Size-1] or 1; 
        end else if rbRSAeSpec.Checked then begin
          OS := HexToOS(edtRSAe.Text);
          if OS = '' then
            raise Exception.Create('Illegal hex value');
          edtRSAe.Text := OSToHex(OS);
          Base256ToUMPInt(IFPubl.E,OS[1],Length(OS),Length(OS)*8);
        end;
        IFRSAKeys(KeySize,2,False,FIFPriv,IFPubl);
      finally
        DisposeIFPublicKey(IFPubl);
      end;

    end else if rbRW.Checked then begin
      KeySize := StrToInt(edtRWKeySize.Text);
      FillChar(IFPubl,SizeOf(IFPubl),0);
      try
        if rbRWeLRN.Checked then begin
          MPRandom(IFPubl.E,KeySize-2);
          MPShl(IFPubl.E,1);
        end else if rbRWeSpec.Checked then begin
          OS := HexToOS(edtRWe.Text);
          if OS = '' then
            raise Exception.Create('Illegal hex value');
          edtRWe.Text := OSToHex(OS);
          Base256ToUMPInt(IFPubl.E,OS[1],Length(OS),Length(OS)*8);
        end;
        IFRWKeys(KeySize,2,False,FIFPriv,IFPubl);
      finally
        DisposeIFPublicKey(IFPubl);
      end;

    end else if rbDL.Checked then begin
      FillChar(DLPubl,SizeOf(DLPubl),0);
      try
        if rbDSA.Checked then
          DLKeys(1024,160,FDLPriv,DLPubl)
        else if rbDLSpecify.Checked then
          DLKeys(StrToInt(edtDLpSize.Text),StrToInt(edtDLqSize.Text),FDLPriv,DLPubl)
        else if rbDLImport.Checked then
          DLKeys(FDLSystemParams,FDLPriv,DLPubl.Y);
      finally
        DisposeDLPublicKey(DLPubl);
      end;
    end;

    btnSave.Enabled := True;
  finally
    sbStatus.Panels[0].Text := 'Done';
  end;
end;

procedure TfrmNewKey.btnSaveClick(Sender: TObject);
var
  FS: TFileStream;
begin
  if SaveDialog1.Execute then begin
    FS := TFileStream.Create(SaveDialog1.FileName,fmCreate);
    try
      if FDLPriv.X = nil then
        RSAPrivateKeyToStream(FS,PW[1],Length(PW),FIFPriv)
      else
        DLPrivateKeyToStream(FS,PW[1],Length(PW),FDLPriv);
    finally
      FS.Free;
    end;
  end;
end;

end.
